package handler

import (
	"bytes"
	"context"
	"encoding/json"
	"log/slog"

	"errors"
	"net/http"
	"net/http/httptest"
	"reflect"
	"testing"

	"gitea.michaelthomson.dev/mthomson/habits/internal/todo/service"
)

type MockTodoCreater struct {
	CreateTodoFunc func(cxt context.Context, todo service.Todo) (service.Todo, error)
}

func (tg *MockTodoCreater) CreateTodo(ctx context.Context, todo service.Todo) (service.Todo, error) {
	return tg.CreateTodoFunc(ctx, todo)
}

func TestCreateTodo(t *testing.T) {
	logger := slog.Default()
	t.Run("create todo", func(t *testing.T) {
		createTodoRequest := CreateTodoRequest{Name: "clean dishes", Done: false}
		createdTodo := service.Todo{Id: 1, Name: "clean dishes", Done: false}
		want := CreateTodoResponse{Id: 1, Name: "clean dishes", Done: false}

		service := MockTodoCreater{
			CreateTodoFunc: func(ctx context.Context, todo service.Todo) (service.Todo, error) {
				return createdTodo, nil
			},
		}

		handler := HandleTodoCreate(logger, &service)

		requestBody, err := json.Marshal(createTodoRequest)

		if err != nil {
			t.Fatalf("Failed to marshal request %+v: %v", createTodoRequest, err)
		}

		req := httptest.NewRequest(http.MethodPost, "/todo", bytes.NewBuffer(requestBody))
		res := httptest.NewRecorder()

		handler(res, req)

		if res.Code != http.StatusCreated {
			t.Errorf("did not get correct status, got %d, want %d", res.Code, http.StatusCreated)
		}

		if res.Header().Get("Location") != "/todo/1" {
			t.Errorf("did not get correct Location, got %q, want %q", res.Header().Get("Location"), "/todo/1")
		}

		var got CreateTodoResponse
		err = json.NewDecoder(res.Body).Decode(&got)

		if err != nil {
			t.Fatalf("Unable to decode response from server %q into GetTodoResponse: %v", res.Body, err)
		}

		if !reflect.DeepEqual(got, want) {
			t.Errorf("got %+v, want %+v", got, want)
		}
	})

	t.Run("returns 400 with bad json", func(t *testing.T) {
		handler := HandleTodoCreate(logger, nil)

		badStruct := struct {
			Foo string
		}{
			Foo: "bar",
		}

		requestBody, err := json.Marshal(badStruct)

		if err != nil {
			t.Fatalf("Failed to marshal request %+v: %v", badStruct, err)
		}

		req := httptest.NewRequest(http.MethodPost, "/todo", bytes.NewBuffer(requestBody))
		res := httptest.NewRecorder()

		handler(res, req)

		if res.Code != http.StatusBadRequest {
			t.Errorf("did not get correct status, got %d, want %d", res.Code, http.StatusOK)
		}
	})

	t.Run("returns 500 arbitrary errors", func(t *testing.T) {
		createTodoRequest := CreateTodoRequest{Name: "clean dishes", Done: false}

		service := MockTodoCreater{
			CreateTodoFunc: func(ctx context.Context, todo service.Todo) (service.Todo, error) {
				return service.Todo{}, errors.New("foo bar")
			},
		}

		handler := HandleTodoCreate(logger, &service)

		requestBody, err := json.Marshal(createTodoRequest)

		if err != nil {
			t.Fatalf("Failed to marshal request %+v: %v", createTodoRequest, err)
		}

		req := httptest.NewRequest(http.MethodPost, "/todo", bytes.NewBuffer(requestBody))
		res := httptest.NewRecorder()

		handler(res, req)

		if res.Code != http.StatusInternalServerError {
			t.Errorf("did not get correct status, got %d, want %d", res.Code, http.StatusOK)
		}
	})

}
