package postgres

import (
	"context"
	"database/sql"
	"log/slog"

	"gitea.michaelthomson.dev/mthomson/habits/internal/todo/repository"
)

type PostgresTodoRepository struct {
	logger *slog.Logger
	db     *sql.DB
}

func NewPostgresTodoRepository(logger *slog.Logger, db *sql.DB) *PostgresTodoRepository {
	return &PostgresTodoRepository{
		logger: logger,
		db:     db,
	}
}

func (r *PostgresTodoRepository) GetById(ctx context.Context, id int64) (repository.TodoRow, error) {
	todo := repository.TodoRow{}

	err := r.db.QueryRow("SELECT * FROM todo WHERE id = $1;", id).Scan(&todo.Id, &todo.Name, &todo.Done)

	if err != nil {
		if err == sql.ErrNoRows {
			return todo, repository.ErrNotFound
		}

		r.logger.ErrorContext(ctx, err.Error())
		return todo, err
	}

	return todo, nil
}

func (r *PostgresTodoRepository) Create(ctx context.Context, todo repository.TodoRow) (repository.TodoRow, error) {
	result := r.db.QueryRow("INSERT INTO todo (name, done) VALUES ($1, $2) RETURNING id;", todo.Name, todo.Done)
	err := result.Scan(&todo.Id)

	if err != nil {
		r.logger.ErrorContext(ctx, err.Error())
		return repository.TodoRow{}, err
	}

	return todo, nil
}

func (r *PostgresTodoRepository) Update(ctx context.Context, todo repository.TodoRow) error {
	result, err := r.db.Exec("UPDATE todo SET name = $1, done = $2 WHERE id = $3;", todo.Name, todo.Done, todo.Id)

	if err != nil {
		r.logger.ErrorContext(ctx, err.Error())
		return err
	}

	rowsAffected, err := result.RowsAffected()

	if err != nil {
		r.logger.ErrorContext(ctx, err.Error())
		return err
	}

	if rowsAffected == 0 {
		return repository.ErrNotFound
	}

	return nil
}

func (r *PostgresTodoRepository) Delete(ctx context.Context, id int64) error {
	result, err := r.db.Exec("DELETE FROM todo WHERE id = $1;", id)

	if err != nil {
		r.logger.ErrorContext(ctx, err.Error())
		return err
	}

	rowsAffected, err := result.RowsAffected()

	if err != nil {
		r.logger.ErrorContext(ctx, err.Error())
		return err
	}

	if rowsAffected == 0 {
		return repository.ErrNotFound
	}

	return nil
}
