package service_test

import (
	"context"
	"log/slog"
	"testing"

	"gitea.michaelthomson.dev/mthomson/habits/internal/test"
	"gitea.michaelthomson.dev/mthomson/habits/internal/todo/repository"
	"gitea.michaelthomson.dev/mthomson/habits/internal/todo/repository/postgres"
	"gitea.michaelthomson.dev/mthomson/habits/internal/todo/service"
	_ "github.com/jackc/pgx/v5/stdlib"
)

func TestCreateTodo(t *testing.T) {
	t.Parallel()
	ctx := context.Background()
	logger := slog.Default()
	tdb := test.NewTestDatabase(t)
	defer tdb.TearDown()
	r := postgres.NewPostgresTodoRepository(logger, tdb.Db)

	todoService := service.NewTodoService(logger, r)

	t.Run("Create todo", func(t *testing.T) {
		todo := service.NewTodo("clean dishes", false)

		_, err := todoService.CreateTodo(ctx, todo)

		AssertNoError(t, err)
	})
}

func TestGetTodo(t *testing.T) {
	t.Parallel()
	ctx := context.Background()
	logger := slog.Default()
	tdb := test.NewTestDatabase(t)
	defer tdb.TearDown()
	r := postgres.NewPostgresTodoRepository(logger, tdb.Db)

	row := repository.TodoRow{Id: 1, Name: "clean dishes", Done: false}
	_, err := r.Create(ctx, row)
	AssertNoError(t, err);

	todoService := service.NewTodoService(logger, r)

	t.Run("Get exisiting todo", func(t *testing.T) {
		_, err := todoService.GetTodo(ctx, 1)

		AssertNoError(t, err)
	})

	t.Run("Get non-existant todo", func(t *testing.T) {
		_, err := todoService.GetTodo(ctx, 2)

		AssertErrors(t, err, service.ErrNotFound)
	})
}

func TestDeleteTodo(t *testing.T) {
	t.Parallel()
	ctx := context.Background()
	logger := slog.Default()
	tdb := test.NewTestDatabase(t)
	defer tdb.TearDown()
	r := postgres.NewPostgresTodoRepository(logger, tdb.Db)

	row := repository.TodoRow{Id: 1, Name: "clean dishes", Done: false}
	_, err := r.Create(ctx, row)
	AssertNoError(t, err);

	todoService := service.NewTodoService(logger, r)

	t.Run("Delete exisiting todo", func(t *testing.T) {
		err := todoService.DeleteTodo(ctx, 1)

		AssertNoError(t, err)
	})

	t.Run("Delete non-existant todo", func(t *testing.T) {
		err := todoService.DeleteTodo(ctx, 1)

		AssertErrors(t, err, service.ErrNotFound)
	})
}

func TestUpdateTodo(t *testing.T) {
	t.Parallel()
	ctx := context.Background()
	logger := slog.Default()
	tdb := test.NewTestDatabase(t)
	defer tdb.TearDown()
	r := postgres.NewPostgresTodoRepository(logger, tdb.Db)

	row := repository.TodoRow{Id: 1, Name: "clean dishes", Done: false}
	_, err := r.Create(ctx, row)
	AssertNoError(t, err);

	todoService := service.NewTodoService(logger, r)

	t.Run("Update exisiting todo", func(t *testing.T) {
		todo := service.Todo{1, "clean dishes", true}

		err := todoService.UpdateTodo(ctx, todo)

		AssertNoError(t, err)

		newTodo, err := todoService.GetTodo(ctx, 1)

		AssertNoError(t, err)

		AssertTodos(t, newTodo, todo)
	})

	t.Run("Update non-existant todo", func(t *testing.T) {
		todo := service.Todo{2, "clean dishes", true}

		err := todoService.UpdateTodo(ctx, todo)

		AssertErrors(t, err, service.ErrNotFound)
	})
}

func AssertErrors(t testing.TB, got, want error) {
	t.Helper()
	if got != want {
		t.Errorf("got error: %v, want error: %v", want, got)
	}
}

func AssertNoError(t testing.TB, err error) {
	t.Helper()
	if err != nil {
		t.Errorf("expected no error, got %v", err)
	}
}

func AssertTodos(t testing.TB, got, want service.Todo) {
	if !got.Equal(want) {
		t.Errorf("got %+v want %+v", got, want)
	}
}
